;; Generated package description from vundo.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "vundo" "2.4.0.0.20260905.15" "Visual undo tree" '((emacs "28.1")) :commit "04caf7ec54e4a7861d3591a1e090adc1fb76b5d1" :authors '(("Yuan Fu" . "casouri@gmail.com")) :maintainer '("Yuan Fu" . "casouri@gmail.com") :keywords '("undo" "text" "editing") :url "https://github.com/casouri/vundo")
